import click

from .commands import add_command, delete_command, list_command, sync_command, update_command


def create_workspace_cli_group():
    """
    Commands for managing a dagster cloud workspace.

    NOTE: currently does not support auth. To work with the K8s demo clouds, you can port forward
    Dagit to your local machine.
    """

    group = click.Group(name="workspace")
    group.add_command(add_command)
    group.add_command(update_command)
    group.add_command(delete_command)
    group.add_command(list_command)
    group.add_command(sync_command)
    return group


workspace_cli = create_workspace_cli_group()



import sys
import threading
import time
from abc import abstractmethod
from typing import Dict, Set

from dagster import check
from dagster.api.get_server_id import sync_get_server_id
from dagster.core.host_representation.grpc_server_registry import GrpcServerRegistry
from dagster.core.host_representation.origin import RegisteredRepositoryLocationOrigin
from dagster.core.host_representation.repository_location import GrpcServerRepositoryLocation
from dagster.core.instance import MayHaveInstanceWeakref
from dagster.daemon.daemon import get_default_daemon_logger
from dagster.grpc.client import DagsterGrpcClient
from dagster.serdes.serdes import serialize_dagster_namedtuple
from dagster.utils.error import SerializableErrorInfo, serializable_error_info_from_exc_info
from dagster_cloud.api.dagster_cloud_api import (
    DagsterCloudRepositoryData,
    DagsterCloudWorkspaceEntry,
    LoadRepositoriesResponse,
)
from dagster_cloud.util import diff_serializable_namedtuple_map
from dagster_cloud.workspace.origin import CodeDeploymentMetadata
from dagster_cloud.workspace.user_code_launcher.query import UPDATE_WORKSPACE_ENTRY_MUTATION

USER_CODE_LAUNCHER_RECONCILE_INTERVAL = 1

UPDATE_WORKSPACE_ENTRY_MUTATION = """
    mutation updateWorkspaceEntryMutation($serializedWorkspaceEntry: String!) {
        workspace {
            updateWorkspaceEntry(serializedWorkspaceEntry: $serializedWorkspaceEntry) {
                ok
            }
        }
    }
"""


class DagsterCloudUserCodeLauncher(GrpcServerRegistry, MayHaveInstanceWeakref):
    def __init__(self, graphql_client=None):
        self._logger = get_default_daemon_logger("DagsterUserCodeLauncher")
        self._graphql_client = graphql_client  # Pulled from the instance outside of tests

    @abstractmethod
    def update_grpc_metadata(
        self, desired_metadata: Dict[str, CodeDeploymentMetadata], force_update_locations: Set[str]
    ):
        pass

    def supports_origin(self, repository_location_origin):
        return isinstance(repository_location_origin, RegisteredRepositoryLocationOrigin)

    @property
    def supports_reload(self):
        return False

    def reload_grpc_endpoint(self, repository_location_origin):
        raise NotImplementedError("Call update_grpc_metadata to update gRPC endpoints")

    @abstractmethod
    def _get_repository_location_origin(self, location_name):
        pass

    def _update_workspace_entry(self, workspace_entry):
        graphql_client = (
            self._graphql_client if self._graphql_client else self._instance.create_graphql_client()
        )
        update_workspace_entry_result = graphql_client.execute(
            UPDATE_WORKSPACE_ENTRY_MUTATION,
            variable_values={
                "serializedWorkspaceEntry": serialize_dagster_namedtuple(
                    check.inst_param(workspace_entry, "workspace_entry", DagsterCloudWorkspaceEntry)
                )
            },
        )
        if "errors" in update_workspace_entry_result:
            raise Exception("Failure updating workspace: " + str(update_workspace_entry_result))

    def _get_repository_load_data(self, repository_location):
        return LoadRepositoriesResponse(
            repository_datas=[
                DagsterCloudRepositoryData(
                    repo_name=external_repository.name,
                    code_pointer=repository_location.repository_code_pointer_dict[
                        external_repository.name
                    ],
                    external_repository_data=external_repository.external_repository_data,
                )
                for external_repository in repository_location.get_repositories().values()
            ],
            container_image=repository_location.container_image,
            executable_path=repository_location.executable_path,
        )

    def _update_location_error(self, location_name, error_info, metadata):
        self._logger.error(
            "Unable to load location {location_name}. Updating location with error data: {error_info}.".format(
                location_name=location_name,
                error_info=str(error_info),
            )
        )

        # Update serialized error
        errored_workspace_entry = DagsterCloudWorkspaceEntry(
            location_name=location_name,
            deployment_metadata=metadata,
            load_repositories_response=None,
            serialized_error_info=error_info,
        )

        self._update_workspace_entry(errored_workspace_entry)

    def _update_location_data(self, location_name, endpoint, metadata):
        self._logger.info(
            "Updating data for location {location_name}".format(location_name=location_name)
        )

        location_origin = self._get_repository_location_origin(location_name)

        if isinstance(endpoint, SerializableErrorInfo):
            self._update_location_error(location_name, error_info=endpoint, metadata=metadata)
            return

        try:
            repository_location = GrpcServerRepositoryLocation(
                origin=location_origin,
                server_id=endpoint.server_id,
                port=endpoint.port,
                socket=endpoint.socket,
                host=endpoint.host,
                heartbeat=True,
                watch_server=False,
                grpc_server_registry=self,
            )
            loaded_workspace_entry = DagsterCloudWorkspaceEntry(
                location_name=location_name,
                deployment_metadata=metadata,
                load_repositories_response=self._get_repository_load_data(repository_location),
                serialized_error_info=None,
            )
        except Exception:  # pylint: disable=broad-except
            self._update_location_error(
                location_name,
                error_info=serializable_error_info_from_exc_info(sys.exc_info()),
                metadata=metadata,
            )
            return

        self._logger.info(
            "Updating location {location_name} with repository load data".format(
                location_name=location_name,
            )
        )
        self._update_workspace_entry(loaded_workspace_entry)


class ReconcileUserCodeLauncher(DagsterCloudUserCodeLauncher):
    def __init__(self, graphql_client=None):
        self._grpc_endpoints = {}
        self._grpc_endpoints_lock = threading.Lock()
        self._logger = get_default_daemon_logger("ReconcileUserCodeLauncher")

        # periodically reconciles to make desired = actual
        self._desired_metadata = {}
        self._actual_metadata = {}
        self._force_update_keys = set()
        self._metadata_lock = threading.Lock()

        super().__init__(graphql_client=graphql_client)

        self._reconcile_count = 0
        self._reconcile_grpc_metadata_shutdown_event = threading.Event()
        self._reconcile_grpc_metadata_thread = threading.Thread(
            target=self._reconcile_thread,
            args=(self._reconcile_grpc_metadata_shutdown_event,),
            name="grpc-reconcile-watch",
        )
        self._reconcile_grpc_metadata_thread.daemon = True
        self._reconcile_grpc_metadata_thread.start()

    def update_grpc_metadata(self, desired_metadata, force_update_locations):
        check.dict_param(
            desired_metadata, "desired_metadata", key_type=str, value_type=CodeDeploymentMetadata
        )
        check.set_param(force_update_locations, "force_update_locations", str)
        with self._metadata_lock:
            self._desired_metadata = desired_metadata
            self._force_update_keys = force_update_locations

    def _get_repository_location_origin(self, location_name):
        return RegisteredRepositoryLocationOrigin(location_name)

    def _reconcile_thread(self, shutdown_event):
        while True:
            shutdown_event.wait(USER_CODE_LAUNCHER_RECONCILE_INTERVAL)
            if shutdown_event.is_set():
                break

            try:
                self.reconcile()
            except Exception:  # pylint: disable=broad-except
                self._logger.error(
                    "Failure updating user code servers: {exc_info}".format(
                        exc_info=sys.exc_info(),
                    )
                )

    def reconcile(self):
        with self._metadata_lock:
            self._reconcile()
            self._reconcile_count += 1

    def _reconcile(self):
        diff = diff_serializable_namedtuple_map(
            self._desired_metadata, self._actual_metadata, self._force_update_keys
        )

        for to_add_key in diff.to_add:
            try:
                new_endpoint = self._add_server(to_add_key, self._desired_metadata[to_add_key])
            except Exception:  # pylint: disable=broad-except
                error_info = serializable_error_info_from_exc_info(sys.exc_info())
                self._logger.error(
                    "Error while adding server for {to_add_key}: {error_info}".format(
                        to_add_key=to_add_key,
                        error_info=error_info,
                    )
                )
                new_endpoint = error_info

            with self._grpc_endpoints_lock:
                self._grpc_endpoints[to_add_key] = new_endpoint

            self._update_location_data(to_add_key, new_endpoint, self._desired_metadata[to_add_key])

            self._actual_metadata[to_add_key] = self._desired_metadata[to_add_key]

        for to_remove_key in diff.to_remove:
            try:
                self._remove_server(to_remove_key, self._actual_metadata[to_remove_key])
            except Exception:  # pylint: disable=broad-except
                self._logger.error(
                    "Error while removing server for {to_remove_key}: {error_info}".format(
                        to_remove_key=to_remove_key,
                        error_info=serializable_error_info_from_exc_info(sys.exc_info()),
                    )
                )

            with self._grpc_endpoints_lock:
                del self._grpc_endpoints[to_remove_key]
            del self._actual_metadata[to_remove_key]

        update_gens = {}
        for to_update_key in diff.to_update:
            update_gens[to_update_key] = self._gen_update_server(
                to_update_key,
                self._actual_metadata[to_update_key],
                self._desired_metadata[to_update_key],
            )
            try:
                new_updated_endpoint = next(update_gens[to_update_key])
            except Exception:  # pylint: disable=broad-except
                error_info = serializable_error_info_from_exc_info(sys.exc_info())
                self._logger.error(
                    "Error while updating server for {to_update_key}: {error_info}".format(
                        to_update_key=to_update_key,
                        error_info=error_info,
                    )
                )
                new_updated_endpoint = error_info

            with self._grpc_endpoints_lock:
                self._grpc_endpoints[to_update_key] = new_updated_endpoint

            self._update_location_data(
                to_update_key, new_updated_endpoint, self._desired_metadata[to_update_key]
            )

        for to_update_key, update_gen in update_gens.items():
            # Finish any remaining cleanup
            try:
                list(update_gen)
            except Exception:  # pylint: disable=broad-except
                self._logger.error(
                    "Error while cleaning up after updating server for {to_update_key}: {error_info}".format(
                        to_update_key=to_update_key,
                        error_info=serializable_error_info_from_exc_info(sys.exc_info()),
                    )
                )

            self._actual_metadata[to_update_key] = self._desired_metadata[to_update_key]
            if to_update_key in self._force_update_keys:
                self._force_update_keys.remove(to_update_key)

    def get_grpc_endpoint(self, repository_location_origin):
        with self._grpc_endpoints_lock:
            location_name = repository_location_origin.location_name
            endpoint = self._grpc_endpoints.get(location_name)

        if not endpoint:
            raise Exception(f"No server endpoint exists for location {location_name}")

        if isinstance(endpoint, SerializableErrorInfo):
            # Consider raising the original exception here instead of a wrapped one
            raise Exception(
                f"Failure loading server endpoint for location {location_name}: {endpoint}"
            )

        return endpoint

    def get_grpc_endpoints(self):
        with self._grpc_endpoints_lock:
            return self._grpc_endpoints.copy()

    @abstractmethod
    def _add_server(self, location_name, metadata):
        # Add a server for this location / metadata combination. This method should be idempotent
        # it's possible if will be called and the server already exists if an update failed
        # part-way through.
        pass

    @abstractmethod
    def _gen_update_server(self, location_name, old_metadata, new_metadata):
        # Update the server for the given location. Is a generator - should yield the new
        # GrpcServerEndpoint, then clean up any no longer needed resources
        pass

    @abstractmethod
    def _remove_server(self, location_name, metadata):
        pass

    def _wait_for_server(self, host, port, timeout=15, socket=None):
        # Wait for the server to be ready (while also loading the server ID)
        server_id = None
        start_time = time.time()
        while True:
            client = DagsterGrpcClient(port=port, host=host, socket=socket)
            try:
                server_id = sync_get_server_id(client)
                break
            except Exception:  # pylint: disable=broad-except
                pass

            if time.time() - start_time > timeout:
                raise Exception(f"Timed out waiting for server {host}:{port}")

            time.sleep(1)
        return server_id

    @abstractmethod
    def step_handler(self):
        pass

    @abstractmethod
    def run_launcher(self):
        pass



UPDATE_WORKSPACE_ENTRY_MUTATION = """
    mutation updateWorkspaceEntryMutation($serializedWorkspaceEntry: String!) {
        workspace {
            updateWorkspaceEntry(serializedWorkspaceEntry: $serializedWorkspaceEntry) {
                ok
            }
        }
    }
"""



import logging
import sys
import time

import pendulum
from dagster import check
from dagster.core.events.log import EventLogEntry
from dagster.core.host_representation import (
    ExternalPipeline,
    ExternalPipelineOrigin,
    PipelineSelector,
)
from dagster.daemon.daemon import DagsterDaemon
from dagster.grpc.types import ExecuteStepArgs
from dagster.serdes import deserialize_json_to_dagster_namedtuple, serialize_dagster_namedtuple
from dagster.utils.error import serializable_error_info_from_exc_info
from dagster.utils.external import external_pipeline_from_location
from dagster_cloud.api.dagster_cloud_api import (
    CheckForWorkspaceUpdatesSuccess,
    DagsterCloudApi,
    DagsterCloudApiErrorResponse,
    DagsterCloudApiSuccess,
)
from dagster_cloud.executor.step_handler_context import DagsterCloudStepHandlerContext

GET_USER_CLOUD_REQUESTS_QUERY = """
    mutation GetUserCloudRequests {
        userCloudAgent {
            popUserCloudAgentRequests(limit:10) {
                requestBody
            }
        }
    }
"""

SEND_USER_CLOUD_RESPONSE_MUTATION = """
    mutation SendUserCloudResponse($requestId: String!, $requestApi: String!, $response: String!) {
        userCloudAgent {
            sendUserCloudAgentResponse(requestId: $requestId, requestApi: $requestApi, responseBody: $response) {
                requestId
            }
        }
    }
"""

WORKSPACE_ENTRIES_QUERY = """
    query WorkspaceEntries {
        workspace {
            workspaceEntries {
                serializedWorkspaceEntry
            }
        }
    }
"""


CLEAR_ERRORS_INTERVAL = 10


class DagsterCloudApiDaemon(DagsterDaemon):
    def __init__(self):
        self._initial_workspace_loaded = False
        super(DagsterCloudApiDaemon, self).__init__(interval_seconds=1)

    @classmethod
    def daemon_type(cls):
        return "DAGSTER_CLOUD_API"

    def __exit__(self, _exception_type, _exception_value, _traceback):
        self._cleanup()

    def __del__(self):
        self._cleanup()

    def _cleanup(self):
        pass

    def _check_for_workspace_updates(self, instance, workspace):
        workspace.cleanup()  # Clear any existing origins

        # Get list of workspace entries from DB
        result = instance.graphql_client.execute(WORKSPACE_ENTRIES_QUERY)
        entries = result["data"]["workspace"]["workspaceEntries"]

        # Create mapping of
        # - location name => deployment metadata
        deployment_map = {}
        force_update_locations = set()
        for entry in entries:
            workspace_entry = deserialize_json_to_dagster_namedtuple(
                entry["serializedWorkspaceEntry"]
            )
            deployment_map[workspace_entry.location_name] = workspace_entry.deployment_metadata
            if workspace_entry.outdated_data:
                force_update_locations.add(workspace_entry.location_name)

        workspace.grpc_server_registry.update_grpc_metadata(deployment_map, force_update_locations)

    def _get_repository_location(self, workspace, repository_location_origin):
        return workspace.get_location(repository_location_origin)

    def _get_external_repository_hierarchy(self, workspace, external_repository_origin):
        repository_location = self._get_repository_location(
            workspace, external_repository_origin.repository_location_origin
        )

        external_repository = repository_location.get_repository(
            external_repository_origin.repository_name
        )
        return (
            repository_location,
            external_repository,
        )

    def _get_external_pipeline_hierarchy(self, workspace, external_pipeline_origin):
        check.inst_param(
            external_pipeline_origin, "external_pipeline_origin", ExternalPipelineOrigin
        )
        external_repository_origin = external_pipeline_origin.external_repository_origin

        repository_location, external_repository = self._get_external_repository_hierarchy(
            workspace, external_repository_origin
        )

        return (
            repository_location,
            external_repository,
            external_repository.get_full_external_pipeline(external_pipeline_origin.pipeline_name),
        )

    def _handle_api_request(self, request, instance, workspace):
        api_name = request.request_api

        if api_name == DagsterCloudApi.CHECK_FOR_WORKSPACE_UPDATES:
            self._check_for_workspace_updates(instance, workspace)
            return CheckForWorkspaceUpdatesSuccess()
        elif api_name == DagsterCloudApi.GET_EXTERNAL_EXECUTION_PLAN:
            external_pipeline_origin = request.request_args.pipeline_origin

            (
                repository_location,
                external_repo,
                external_pipeline,
            ) = self._get_external_pipeline_hierarchy(workspace, external_pipeline_origin)

            if request.request_args.solid_selection:
                selector = PipelineSelector(
                    repository_location.name,
                    external_repo.name,
                    external_pipeline.name,
                    request.request_args.solid_selection,
                )
                subset_result = repository_location.get_subset_external_pipeline_result(selector)
                external_pipeline = ExternalPipeline(
                    subset_result.external_pipeline_data,
                    external_repo.handle,
                )

            return repository_location.get_external_execution_plan(
                external_pipeline=external_pipeline,
                run_config=request.request_args.run_config,
                mode=request.request_args.mode,
                step_keys_to_execute=request.request_args.step_keys_to_execute,
                known_state=request.request_args.known_state,
            ).execution_plan_snapshot

        elif api_name == DagsterCloudApi.GET_SUBSET_EXTERNAL_PIPELINE_RESULT:
            external_pipeline_origin = request.request_args.pipeline_origin
            (
                repository_location,
                external_repository,
                external_pipeline,
            ) = self._get_external_pipeline_hierarchy(workspace, external_pipeline_origin)
            return repository_location.get_subset_external_pipeline_result(
                PipelineSelector(
                    location_name=repository_location.name,
                    repository_name=external_repository.name,
                    pipeline_name=external_pipeline.name,
                    solid_selection=request.request_args.solid_selection,
                )
            )
        elif api_name == DagsterCloudApi.GET_EXTERNAL_PARTITION_CONFIG:
            external_repository_origin = request.request_args.repository_origin
            repository_location, external_repository = self._get_external_repository_hierarchy(
                workspace, external_repository_origin
            )
            return repository_location.get_external_partition_config(
                external_repository.handle,
                request.request_args.partition_set_name,
                request.request_args.partition_name,
            )
        elif api_name == DagsterCloudApi.GET_EXTERNAL_PARTITION_TAGS:
            external_repository_origin = request.request_args.repository_origin
            repository_location, external_repository = self._get_external_repository_hierarchy(
                workspace, external_repository_origin
            )
            return repository_location.get_external_partition_tags(
                external_repository.handle,
                request.request_args.partition_set_name,
                request.request_args.partition_name,
            )
        elif api_name == DagsterCloudApi.GET_EXTERNAL_PARTITION_NAMES:
            external_repository_origin = request.request_args.repository_origin
            repository_location, external_repository = self._get_external_repository_hierarchy(
                workspace, external_repository_origin
            )

            return repository_location.get_external_partition_names(
                external_repository.handle, request.request_args.partition_set_name
            )
        elif api_name == DagsterCloudApi.GET_EXTERNAL_PARTITION_SET_EXECUTION_PARAM_DATA:
            external_repository_origin = request.request_args.repository_origin
            repository_location, external_repository = self._get_external_repository_hierarchy(
                workspace, external_repository_origin
            )
            return repository_location.get_external_partition_set_execution_param_data(
                external_repository.handle,
                request.request_args.partition_set_name,
                request.request_args.partition_names,
            )
        elif api_name == DagsterCloudApi.GET_EXTERNAL_SCHEDULE_EXECUTION_DATA:
            external_repository_origin = request.request_args.repository_origin

            repository_location, external_repository = self._get_external_repository_hierarchy(
                workspace, external_repository_origin
            )
            scheduled_execution_time = (
                pendulum.from_timestamp(
                    request.request_args.scheduled_execution_timestamp,
                    tz=request.request_args.scheduled_execution_timezone,
                )
                if request.request_args.scheduled_execution_timestamp
                else None
            )
            return repository_location.get_external_schedule_execution_data(
                instance,
                external_repository.handle,
                request.request_args.schedule_name,
                scheduled_execution_time,
            )
        elif api_name == DagsterCloudApi.GET_EXTERNAL_SENSOR_EXECUTION_DATA:
            external_repository_origin = request.request_args.repository_origin

            repository_location, external_repository = self._get_external_repository_hierarchy(
                workspace, external_repository_origin
            )
            return repository_location.get_external_sensor_execution_data(
                instance,
                external_repository.handle,
                request.request_args.sensor_name,
                request.request_args.last_completion_time,
                request.request_args.last_run_key,
                request.request_args.cursor,
            )
        elif api_name == DagsterCloudApi.LAUNCH_RUN:
            run = request.request_args.pipeline_run
            instance.report_engine_event(
                f"Received request from {instance.dagster_cloud_url} to launch all steps for pipeline {run.pipeline_name}",
                run,
                cls=self.__class__,
            )
            repository_location = self._get_repository_location(
                workspace,
                run.external_pipeline_origin.external_repository_origin.repository_location_origin,
            )

            external_pipeline = external_pipeline_from_location(
                repository_location, run.external_pipeline_origin, run.solid_selection
            )

            launcher = workspace.grpc_server_registry.run_launcher()
            launcher.launch_run(run, external_pipeline)
            return DagsterCloudApiSuccess()
        elif api_name == DagsterCloudApi.LAUNCH_STEP:
            context = DagsterCloudStepHandlerContext.deserialize(
                instance, request.request_args.persisted_step_handler_context
            )
            args: ExecuteStepArgs = context.execute_step_args
            assert len(args.step_keys_to_execute) == 1
            step_key = args.step_keys_to_execute[0]

            instance.report_engine_event(
                f"Received request from {instance.dagster_cloud_url} to launch steps: {', '.join(args.step_keys_to_execute)}",
                context.pipeline_run,
                cls=self.__class__,
            )

            step_handler = workspace.grpc_server_registry.step_handler()
            events = step_handler.launch_step(context)
            for event in events:
                event_record = EventLogEntry(
                    message=event.message,
                    user_message=event.message,
                    level=logging.INFO,
                    pipeline_name=context.pipeline_run.pipeline_name,
                    run_id=context.pipeline_run.run_id,
                    error_info=None,
                    timestamp=time.time(),
                    step_key=step_key,
                    dagster_event=event,
                )
                instance.handle_new_event(event_record)

            return DagsterCloudApiSuccess()

        elif api_name == DagsterCloudApi.TERMINATE_STEP:
            context = DagsterCloudStepHandlerContext.deserialize(
                instance, request.request_args.persisted_step_handler_context
            )

            step_handler = workspace.grpc_server_registry.step_handler()
            events = step_handler.terminate_step(context)
            for event in events:
                instance.handle_new_event(event)

            return DagsterCloudApiSuccess()

        elif api_name == DagsterCloudApi.CHECK_STEP_HEALTH:
            context = DagsterCloudStepHandlerContext.deserialize(
                instance, request.request_args.persisted_step_handler_context
            )

            step_handler = workspace.grpc_server_registry.step_handler()
            events = step_handler.check_step_health(context)
            for event in events:
                instance.handle_new_event(event)

            return DagsterCloudApiSuccess()

        else:
            raise Exception(
                "Unexpected dagster cloud api call {api_name}".format(api_name=api_name)
            )

    def run_iteration(self, instance, workspace):
        if not self._initial_workspace_loaded:
            self._check_for_workspace_updates(instance, workspace)
            self._initial_workspace_loaded = True

        while True:
            result = instance.graphql_client.execute(GET_USER_CLOUD_REQUESTS_QUERY)

            requests = [
                deserialize_json_to_dagster_namedtuple(request["requestBody"])
                for request in result["data"]["userCloudAgent"]["popUserCloudAgentRequests"]
            ]

            for request in requests:
                self._logger.info(
                    "Responding to API request {request_id}: {request_name} with args: {request_args}".format(
                        request_id=request.request_id,
                        request_name=request.request_api,
                        request_args=str(request.request_args) if request.request_args else "None",
                    )
                )

                api_result = None
                try:
                    api_result = self._handle_api_request(request, instance, workspace)
                except Exception:  # pylint: disable=broad-except
                    error_info = serializable_error_info_from_exc_info(sys.exc_info())
                    api_result = DagsterCloudApiErrorResponse(error_infos=[error_info])
                    self._logger.error(
                        "Error serving API request {request_id}: {request_name}:\n{error_info}".format(
                            request_id=request.request_id,
                            request_name=request.request_api,
                            error_info=str(error_info),
                        )
                    )
                    yield error_info

                response = serialize_dagster_namedtuple(api_result)

                # Check for errors
                instance.graphql_client.execute(
                    SEND_USER_CLOUD_RESPONSE_MUTATION,
                    {
                        "requestId": request.request_id,
                        "requestApi": request.request_api.value,
                        "response": response,
                    },
                )

            time.sleep(0.5)
            yield



import os
import sys
import uuid
from typing import List

import docker
from dagster import Array, Field, check
from dagster.core.events import DagsterEvent
from dagster.core.executor.step_delegating.step_handler.base import StepHandler, StepHandlerContext
from dagster.core.host_representation.grpc_server_registry import GrpcServerEndpoint
from dagster.core.launcher import DefaultRunLauncher
from dagster.daemon.daemon import get_default_daemon_logger
from dagster.serdes import ConfigurableClass, serialize_dagster_namedtuple
from dagster.serdes.utils import create_snapshot_id
from dagster.utils import find_free_port, merge_dicts

from .user_code_launcher import ReconcileUserCodeLauncher

GRPC_SERVER_LABEL = "dagster_grpc_server"


class DockerUserCodeLauncher(ReconcileUserCodeLauncher, ConfigurableClass):
    def __init__(
        self,
        inst_data=None,
        networks=None,
        env_vars=None,
    ):
        self._inst_data = inst_data
        self._logger = get_default_daemon_logger("DockerUserCodeLauncher")
        self._networks = check.opt_list_param(networks, "networks", of_type=str)
        self._env_vars = env_vars

        self._cleanup_containers()
        super(DockerUserCodeLauncher, self).__init__()

        self._launcher = DefaultRunLauncher()

    def register_instance(self, instance):
        super().register_instance(instance)
        self._launcher.register_instance(instance)

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {
            "networks": Field(Array(str), is_required=False),
            "env_vars": Field(
                [str],
                is_required=False,
                description="The list of environment variables names to forward to the docker container",
            ),
        }

    @staticmethod
    def from_config_value(inst_data, config_value):
        return DockerUserCodeLauncher(inst_data=inst_data, **config_value)

    def _create_container(self, client, location_name, metadata, container_name, hostname, port):

        python_env = (
            {"DAGSTER_CLI_API_GRPC_PYTHON_FILE": metadata.python_file}
            if metadata.python_file
            else {"DAGSTER_CLI_API_GRPC_PACKAGE_NAME": metadata.package_name}
        )
        return client.containers.create(
            metadata.image,
            detach=True,
            hostname=hostname,
            name=container_name,
            network=self._networks[0] if len(self._networks) else None,
            environment=merge_dicts(
                (
                    {env_name: os.getenv(env_name) for env_name in self._env_vars}
                    if self._env_vars
                    else {}
                ),
                {
                    "DAGSTER_CURRENT_IMAGE": metadata.image,
                    "PYTHONUNBUFFERED": "1",
                    "DAGSTER_CLI_API_GRPC_PORT": str(port),
                    "DAGSTER_CLI_API_GRPC_HOST": "0.0.0.0",
                    "DAGSTER_CLI_API_GRPC_LAZY_LOAD_USER_CODE": "1",
                },
                python_env,
            ),
            labels=[GRPC_SERVER_LABEL, self._get_label_for_metadata(location_name, metadata)],
            command=["dagster", "api", "grpc"],
            ports={port: port} if hostname == "localhost" else None,
        )

    def _get_containers(self, client, location_name, metadata):
        return client.containers.list(
            all=True, filters={"label": self._get_label_for_metadata(location_name, metadata)}
        )

    def _add_server(self, location_name, metadata):
        client = docker.client.from_env()

        # check if the container already exists, remove it if so (could happen
        # if a previous attempt to add the container failed)
        existing_containers = self._get_containers(client, location_name, metadata)
        for existing_container in existing_containers:
            self._logger.info(
                "Removing existing container for location {location_name} with image {image}: {container_name}".format(
                    location_name=location_name,
                    image=metadata.image,
                    container_name=existing_container.name,
                )
            )
            self._remove_container(existing_container)

        return self._start_new_server(location_name, metadata)

    def _start_new_server(self, location_name, metadata):
        client = docker.client.from_env()

        container_name = f"{location_name}_{str(uuid.uuid4().hex)[0:6]}"

        self._logger.info(
            "Starting a new container for location {location_name} with image {image}: {container_name}".format(
                location_name=location_name, image=metadata.image, container_name=container_name
            )
        )

        has_network = len(self._networks) > 0
        if has_network:
            port = 4000
            hostname = container_name
        else:
            port = find_free_port()
            hostname = "localhost"

        try:
            container = self._create_container(
                client, location_name, metadata, container_name, hostname, port
            )
        except docker.errors.ImageNotFound:
            client.images.pull(metadata.image)
            container = self._create_container(
                client, location_name, metadata, container_name, hostname, port
            )

        if len(self._networks) > 1:
            for network_name in self._networks[1:]:
                network = client.networks.get(network_name)
                network.connect(container)

        container.start()

        self._logger.info("Started container {container_id}".format(container_id=container.id))

        server_id = self._wait_for_server(host=hostname, port=port)

        endpoint = GrpcServerEndpoint(
            server_id=server_id,
            host=hostname,
            port=port,
            socket=None,
        )

        return endpoint

    def _gen_update_server(self, location_name, old_metadata, new_metadata):
        client = docker.client.from_env()
        existing_containers = self._get_containers(client, location_name, old_metadata)
        updated_server = self._start_new_server(location_name, new_metadata)
        yield updated_server

        # Now cleanup the old containers
        for existing_container in existing_containers:
            self._logger.info(
                "Removing old container for location {location_name}: {container_name}".format(
                    location_name=location_name,
                    container_name=existing_container.name,
                )
            )
            self._remove_container(existing_container)

    def _get_label_for_metadata(self, location_name, metadata):
        return f"{location_name}_{create_snapshot_id(metadata)[0:6]}"

    def _remove_server(self, location_name, metadata):
        client = docker.client.from_env()
        containers = self._get_containers(client, location_name, metadata)
        for container in containers:
            self._logger.info(
                "Removing a container for location {location_name} with image {image}: {container_name}".format(
                    location_name=location_name,
                    image=container.image,
                    container_name=container.name,
                )
            )

            self._remove_container(container)

    def _remove_container(self, container):
        try:
            container.stop()
        except Exception:  # pylint: disable=broad-except
            self._logger.error(
                "Failure stopping container {container_id}: {exc_info}".format(
                    container_id=container.id,
                    exc_info=sys.exc_info(),
                )
            )
        container.remove(force=True)
        self._logger.info("Removed container {container_id}".format(container_id=container.id))

    def _cleanup_containers(self):
        client = docker.client.from_env()
        # Shut down any dangling containers with the gRPC label
        existing_containers = client.containers.list(all=True, filters={"label": GRPC_SERVER_LABEL})

        for container in existing_containers:
            self._logger.info(
                "Stopping and removing existing gRPC container {container_name}".format(
                    container_name=container.name
                )
            )
            self._remove_container(container)

    def __exit__(self, exception_type, exception_value, traceback):
        self._launcher.dispose()
        self._cleanup_containers()

    def step_handler(self):
        return DockerStepHandler(self._networks, self._env_vars)

    def run_launcher(self):
        return self._launcher


class DockerStepHandler(StepHandler):
    def __init__(self, networks, env_vars):
        super().__init__()
        self._networks = check.opt_list_param(networks, "networks", of_type=str)
        self._env_vars = check.opt_list_param(env_vars, "env_vars", of_type=str)

    @property
    def name(self) -> str:
        return "DockerStepHandler"

    def _create_step_container(self, client, step_image, execute_step_args):

        return client.containers.create(
            step_image,
            detach=True,
            network=self._networks[0] if len(self._networks) else None,
            command=[
                "dagster",
                "api",
                "execute_step",
                serialize_dagster_namedtuple(execute_step_args),
            ],
            environment=(
                {env_name: os.getenv(env_name) for env_name in self._env_vars}
                if self._env_vars
                else {}
            ),
        )

    def launch_step(self, step_handler_context: StepHandlerContext) -> List[DagsterEvent]:
        client = docker.client.from_env()

        step_image = (
            step_handler_context.execute_step_args.pipeline_origin.repository_origin.container_image
        )

        if not step_image:
            raise Exception("No image included to launch steps: " + str(step_image))

        try:
            step_container = self._create_step_container(
                client, step_image, step_handler_context.execute_step_args
            )
        except docker.errors.ImageNotFound:
            client.images.pull(step_image)
            step_container = self._create_step_container(
                client, step_image, step_handler_context.execute_step_args
            )

        if len(self._networks) > 1:
            for network_name in self._networks[1:]:
                network = client.networks.get(network_name)
                network.connect(step_container)

        step_container.start()
        return []

    def check_step_health(self, step_handler_context: StepHandlerContext) -> List[DagsterEvent]:
        # TODO not implemented
        return []

    def terminate_step(self, step_handler_context: StepHandlerContext) -> List[DagsterEvent]:
        # TODO not implemented
        return []



import os
import time

import click
import yaml
from dagster import check
from dagster.serdes.serdes import deserialize_json_to_dagster_namedtuple
from dagster.utils import DEFAULT_WORKSPACE_YAML_FILENAME
from dagster_cloud.auth.constants import API_TOKEN_HEADER
from dagster_cloud.storage.client import GqlShimClient
from dagster_cloud.workspace.config_schema import process_workspace_config
from dagster_cloud.workspace.origin import CodeDeploymentMetadata
from gql import Client
from gql.transport.requests import RequestsHTTPTransport


def add_options(options):
    """
    Share options between commands
    https://stackoverflow.com/questions/40182157/shared-options-and-flags-between-commands
    """

    def _add_options(func):
        for option in reversed(options):
            func = option(func)
        return func

    return _add_options


DEFAULT_DAGIT_URL = lambda: os.getenv(
    "DAGSTER_CLOUD_DAGIT_URL",
    "http://localhost:3000",
)


def _api_token():
    return os.getenv("DAGSTER_CLOUD_API_TOKEN")


def _show_api_token_prompt():
    password_exists = bool(_api_token())
    show_prompt = not password_exists
    return show_prompt


_dagit_client_options = [
    click.option("--url", type=click.STRING, default=DEFAULT_DAGIT_URL, help="Dagit url"),
    click.option(
        "--api-token",
        type=click.STRING,
        default=_api_token,
        prompt=_show_api_token_prompt(),
        hide_input=True,
        required=True,
        help="API token generated in Dagit",
    ),
]

_deployment_metadata_options = [
    click.option("--image", type=click.STRING, help="Docker image"),
    click.option(
        "--python-file",
        "-f",
        type=click.Path(exists=False),
        help="python file where repository lives",
    ),
    click.option(
        "--package-name",
        "-p",
        help="Python package where repositories live",
    ),
]


def _dagit_client(url, api_token):
    url = f"{url}/graphql"

    transport = RequestsHTTPTransport(
        url=url,
        use_json=True,
        headers={"Content-type": "application/json", API_TOKEN_HEADER: api_token},
        retries=3,
    )
    return _graphql_client(transport)


def _graphql_client(transport):
    return GqlShimClient(client=Client(transport=transport, fetch_schema_from_transport=True))


ADD_LOCATION_MUTATION = """
mutation ($location: LocationSelector!) {
   addLocation(location: $location) {
      __typename
      ... on WorkspaceEntry {
        locationName
      }
      ... on PythonError {
        message
        stack
      }
   }
}
"""

REPOSITORY_LOCATIONS_QUERY = """
{
  workspaceOrError {
    __typename
    ... on Workspace {
      locationEntries {
        __typename
        name
        loadStatus
        locationOrLoadError {
          __typename
          ... on RepositoryLocation {
            name
          }
          ... on PythonError {
              message
              stack
          }
        }
      }
    }
    ... on PythonError {
      message
      stack
    }
  }
}
"""


@click.command(
    name="add-location",
    help="Add a new repository location to the workspace.",
)
@add_options(_dagit_client_options)
@add_options(_deployment_metadata_options)
@click.argument("location", type=click.STRING)
def add_command(url, api_token, location, **kwargs):
    execute_add_command(_dagit_client(url, api_token), location, kwargs)


def _get_location_input(location, kwargs):
    location_input = {"name": location}

    python_file = kwargs.get("python_file")
    package_name = kwargs.get("package_name")

    if (not python_file and not package_name) or (python_file and package_name):
        raise click.UsageError("Must specify exactly one of --python-file or --package-name.")

    if python_file:
        location_input["pythonFile"] = python_file
    if package_name:
        location_input["packageName"] = package_name
    if "image" in kwargs:
        location_input["image"] = kwargs["image"]

    return location_input


def execute_add_command(client, location, kwargs):
    variables = {"location": _get_location_input(location, kwargs)}

    click.echo(f"Adding location {location}...")

    add_res = client.execute(ADD_LOCATION_MUTATION, variable_values=variables)

    if add_res["data"]["addLocation"]["__typename"] == "WorkspaceEntry":
        click.echo(f"Added location {location}.")
    else:
        raise click.ClickException(f"Error adding location: {str(add_res)}")

    wait_for_load(client, [location])


def wait_for_load(client, locations):
    start_time = time.time()
    click.echo(f"Waiting for agent to sync changes to {','.join(locations)}...")
    while True:
        if time.time() - start_time > 180:
            raise click.ClickException("Timed out waiting for location data to update")

        repo_locations_res = client.execute(REPOSITORY_LOCATIONS_QUERY)

        if not "data" in repo_locations_res:
            raise click.ClickException(f"Error fetching locations: {str(repo_locations_res)}")
        locations_or_error = repo_locations_res["data"]["workspaceOrError"]
        if locations_or_error["__typename"] == "PythonError" or not "data":
            raise click.ClickException(f"Error fetching locations: {str(locations_or_error)}")

        nodes = repo_locations_res["data"]["workspaceOrError"]["locationEntries"]

        nodes_by_location = {node["name"]: node for node in nodes}

        if all(
            location in nodes_by_location
            and nodes_by_location[location].get("loadStatus") == "LOADED"
            for location in locations
        ):

            error_locations = [
                location
                for location in locations
                if "locationOrLoadError" in nodes_by_location[location]
                and nodes_by_location[location]["locationOrLoadError"]["__typename"]
                == "PythonError"
            ]

            if error_locations:
                error_string = "Some locations failed to load after being synced by the agent:\n" + "\n".join(
                    [
                        f"Error loading {error_location}: {str(nodes_by_location[error_location]['locationOrLoadError'])}"
                        for error_location in error_locations
                    ]
                )
                raise click.ClickException(error_string)
            else:
                click.echo(
                    f"Agent synced changes to {','.join(locations)}. Changes should now be visible in dagit."
                )
                break

        time.sleep(3)


UPDATE_LOCATION_MUTATION = """
mutation ($location: LocationSelector!) {
   updateLocation(location: $location) {
      __typename
      ... on WorkspaceEntry {
        locationName
      }
      ... on PythonError {
        message
        stack
      }
   }
}
"""


@click.command(
    name="update-location",
    help="Update the image for a repository location in the workspace.",
)
@add_options(_dagit_client_options)
@add_options(_deployment_metadata_options)
@click.argument("location", type=click.STRING)
def update_command(url, api_token, location, **kwargs):
    execute_update_command(_dagit_client(url, api_token), location, kwargs)


def execute_update_command(client, location, kwargs):
    click.echo(f"Updating location {location}...")

    variables = {"location": _get_location_input(location, kwargs)}

    update_res = client.execute(UPDATE_LOCATION_MUTATION, variable_values=variables)

    if update_res["data"]["updateLocation"]["__typename"] == "WorkspaceEntry":
        click.echo(f"Updated location {location}.")
    else:
        raise click.ClickException(f"Error adding location: {str(update_res)}")

    wait_for_load(client, [location])


DELETE_LOCATION_MUTATION = """
mutation ($locationName: String!) {
   deleteLocation(locationName: $locationName) {
      __typename
      ... on DeleteLocationSuccess {
        locationName
      }
      ... on PythonError {
        message
        stack
      }
   }
}
"""


@click.command(
    name="delete-location",
    help="Delete a repository location from the workspace.",
)
@add_options(_dagit_client_options)
@click.argument("location", type=click.STRING)
def delete_command(url, api_token, location):
    execute_delete_command(_dagit_client(url, api_token), location)


def execute_delete_command(client, location):
    click.echo(f"Deleting location {location}...")

    variables = {"locationName": location}

    delete_res = client.execute(DELETE_LOCATION_MUTATION, variable_values=variables)

    if delete_res["data"]["deleteLocation"]["__typename"] == "DeleteLocationSuccess":
        click.echo(f"Deleted location {location}")
    else:
        raise click.ClickException(f"Error deleting location: {str(delete_res)}")


LIST_LOCATIONS_QUERY = """
query WorkspaceEntries {
    workspace {
        workspaceEntries {
            locationName
            serializedDeploymentMetadata
        }
    }
}
"""


@click.command(
    name="list",
    help="List repository locations in the workspace.",
)
@add_options(_dagit_client_options)
def list_command(url, api_token):
    execute_list_command(_dagit_client(url, api_token))


def execute_list_command(client):
    click.echo("Listing locations...")

    list_res = client.execute(LIST_LOCATIONS_QUERY)

    for location in list_res["data"]["workspace"]["workspaceEntries"]:
        metadata = check.inst(
            deserialize_json_to_dagster_namedtuple(location["serializedDeploymentMetadata"]),
            CodeDeploymentMetadata,
        )

        location_desc = [location["locationName"]]
        if metadata.python_file:
            location_desc.append(f"File: {metadata.python_file}")
        if metadata.package_name:
            location_desc.append(f"Package: {metadata.package_name}")
        if metadata.image:
            location_desc.append(f"Image: {metadata.image}")

        click.echo("\t".join(location_desc))


RECONCILE_LOCATIONS_MUTATION = """
mutation ($locations: [LocationSelector]!) {
    reconcileLocations(locations: $locations) {
        __typename
        ... on ReconcileLocationsSuccess {
            locations {
                locationName
            }
        }
        ... on PythonError {
            message
            stack
        }
    }
}
"""


@click.command(name="sync", help="Sync the workspace with the contents of a workspace.yaml file.")
@add_options(_dagit_client_options)
@click.option(
    "--workspace",
    "-w",
    type=click.Path(exists=True),
    default=DEFAULT_WORKSPACE_YAML_FILENAME,
    help="Path to workspace file.",
)
def sync_command(url, api_token, workspace):
    execute_sync_command(_dagit_client(url, api_token), workspace)


def execute_sync_command(client, workspace):
    with open(str(workspace), "r") as f:
        config = yaml.load(f.read(), Loader=yaml.SafeLoader)
        process_workspace_config(config)

    locations = {"locations": []}
    for name, metadata in config["locations"].items():
        locations["locations"].append(
            {
                "name": name,
                "image": metadata.get("image"),
                "packageName": metadata.get("package_name"),
                "pythonFile": metadata.get("python_file"),
            }
        )

    reconcile_res = client.execute(RECONCILE_LOCATIONS_MUTATION, variable_values=locations)

    if reconcile_res["data"]["reconcileLocations"]["__typename"] == "ReconcileLocationsSuccess":
        locations = sorted(
            [
                location["locationName"]
                for location in reconcile_res["data"]["reconcileLocations"]["locations"]
            ]
        )
        click.echo(f"Synced locations: {', '.join(locations)}")
    else:
        raise click.ClickException(f"Error syncing locations: {str(reconcile_res)}")

    wait_for_load(client, locations)



from typing import Iterable

from dagster import check
from dagster.core.definitions.run_request import JobType
from dagster.core.scheduler.job import (
    JobState,
    JobTick,
    JobTickData,
    JobTickStatsSnapshot,
    JobTickStatus,
)
from dagster.core.storage.schedules.base import ScheduleStorage
from dagster.serdes import (
    ConfigurableClass,
    ConfigurableClassData,
    deserialize_json_to_dagster_namedtuple,
    serialize_dagster_namedtuple,
)

from ..errors import GraphQLStorageError
from .queries import (
    ADD_JOB_STATE_MUTATION,
    ALL_STORED_JOB_STATE_QUERY,
    CREATE_JOB_TICK_MUTATION,
    GET_JOB_STATE_QUERY,
    GET_JOB_TICKS_QUERY,
    GET_JOB_TICK_STATS_QUERY,
    LATEST_JOB_TICK_QUERY,
    PURGE_JOB_TICKS_MUTATION,
    UPDATE_JOB_STATE_MUTATION,
    UPDATE_JOB_TICK_MUTATION,
)


class GraphQLScheduleStorage(ScheduleStorage, ConfigurableClass):
    def __init__(self, inst_data=None, override_graphql_client=None):
        """Initialize this class directly only for test (using `override_graphql_client`).
        Use the ConfigurableClass machinery to init from instance yaml."""
        self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)
        self._override_graphql_client = override_graphql_client

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {}

    @staticmethod
    def from_config_value(inst_data: ConfigurableClassData, config_value):
        return GraphQLScheduleStorage(inst_data=inst_data)

    @property
    def _graphql_client(self):
        return (
            self._override_graphql_client
            if self._override_graphql_client
            else self._instance.graphql_client
        )

    def _execute_query(self, query, variables=None):
        res = self._graphql_client.execute(query, variable_values=variables)
        if "errors" in res:
            raise GraphQLStorageError(res)
        return res

    def wipe(self):
        raise Exception("Not allowed to wipe from user cloud")

    def all_stored_job_state(
        self, repository_origin_id: str = None, job_type: JobType = None
    ) -> Iterable[JobState]:
        res = self._execute_query(
            ALL_STORED_JOB_STATE_QUERY,
            variables={
                "repositoryOriginId": repository_origin_id,
                "jobType": job_type.value if job_type else None,
            },
        )

        return [
            check.inst(deserialize_json_to_dagster_namedtuple(job_state), JobState)
            for job_state in res["data"]["schedules"]["jobStates"]
        ]

    def get_job_state(self, job_origin_id: str) -> JobState:
        res = self._execute_query(
            GET_JOB_STATE_QUERY,
            variables={
                "jobOriginId": job_origin_id,
            },
        )

        job_state = res["data"]["schedules"]["jobState"]
        if job_state is None:
            return None

        return check.inst(deserialize_json_to_dagster_namedtuple(job_state), JobState)

    def add_job_state(self, job: JobState):
        self._execute_query(
            ADD_JOB_STATE_MUTATION,
            variables={
                "serializedJobState": serialize_dagster_namedtuple(
                    check.inst_param(job, "job", JobState)
                )
            },
        )

    def update_job_state(self, job: JobState):
        self._execute_query(
            UPDATE_JOB_STATE_MUTATION,
            variables={
                "serializedJobState": serialize_dagster_namedtuple(
                    check.inst_param(job, "job", JobState)
                )
            },
        )

    def delete_job_state(self, job_origin_id: str):
        raise NotImplementedError("Not callable from user cloud")

    def get_job_ticks(
        self, job_origin_id: str, before: float = None, after: float = None, limit: int = None
    ) -> Iterable[JobTick]:
        res = self._execute_query(
            GET_JOB_TICKS_QUERY,
            variables={
                "jobOriginId": job_origin_id,
                "before": before,
                "after": after,
                "limit": limit,
            },
        )

        return [
            check.inst(deserialize_json_to_dagster_namedtuple(job_tick), JobTick)
            for job_tick in res["data"]["schedules"]["jobTicks"]
        ]

    def get_latest_job_tick(self, job_origin_id: str) -> JobTick:
        res = self._execute_query(
            LATEST_JOB_TICK_QUERY,
            variables={
                "jobOriginId": job_origin_id,
            },
        )

        latest_tick = res["data"]["schedules"]["latestJobTick"]
        if not latest_tick:
            return None

        return check.inst(deserialize_json_to_dagster_namedtuple(latest_tick), JobTick)

    def create_job_tick(self, job_tick_data: JobTickData):
        res = self._execute_query(
            CREATE_JOB_TICK_MUTATION,
            variables={
                "serializedJobTickData": serialize_dagster_namedtuple(
                    check.inst_param(job_tick_data, "job_tick_data", JobTickData)
                )
            },
        )

        tick_id = res["data"]["schedules"]["createJobTick"]["tickId"]
        return JobTick(tick_id, job_tick_data)

    def update_job_tick(self, tick: JobTick):
        self._execute_query(
            UPDATE_JOB_TICK_MUTATION,
            variables={
                "tickId": tick.tick_id,
                "serializedJobTickData": serialize_dagster_namedtuple(
                    check.inst_param(tick.job_tick_data, "job_tick_data", JobTickData)
                ),
            },
        )
        return tick

    def purge_job_ticks(self, job_origin_id: str, tick_status: JobTickStatus, before: float):
        self._execute_query(
            PURGE_JOB_TICKS_MUTATION,
            variables={
                "jobOriginId": job_origin_id,
                "tickStatus": tick_status.value if tick_status else None,
                "before": before,
            },
        )

    def get_job_tick_stats(self, job_origin_id: str):
        res = self._execute_query(
            GET_JOB_TICK_STATS_QUERY,
            variables={
                "jobOriginId": job_origin_id,
            },
        )

        stats = res["data"]["schedules"]["jobTickStats"]
        return JobTickStatsSnapshot(
            ticks_started=stats["ticksStarted"],
            ticks_succeeded=stats["ticksSucceeded"],
            ticks_skipped=stats["ticksSkipped"],
            ticks_failed=stats["ticksFailed"],
        )

    def upgrade(self):
        raise NotImplementedError("Not callable from user cloud")

    def optimize_for_dagit(self, statement_timeout: int):
        raise NotImplementedError("Not callable from user cloud")



from collections import namedtuple
from datetime import timedelta
from enum import Enum

import pendulum
from dagster import check
from dagster.core.code_pointer import CodePointer
from dagster.core.host_representation import ExternalRepositoryData, RepositoryLocationOrigin
from dagster.core.storage.pipeline_run import PipelineRun
from dagster.serdes import whitelist_for_serdes
from dagster.utils.error import SerializableErrorInfo
from dagster_cloud.executor.step_handler_context import PersistedDagsterCloudStepHandlerContext
from dagster_cloud.workspace.origin import CodeDeploymentMetadata

DEFAULT_EXPIRATION_MILLISECONDS = 10 * 60 * 1000


@whitelist_for_serdes
class DagsterCloudWorkspaceEntryStatus(Enum):
    LOADING = "LOADING"
    LOADED = "LOADED"
    ERRORED = "ERRORED"


@whitelist_for_serdes
class DagsterCloudWorkspaceEntry(
    namedtuple(
        "_DagsterCloudWorkspaceEntry",
        "location_name deployment_metadata load_repositories_response serialized_error_info timestamp data_updated_timestamp",
    )
):
    """Includes information to tell the daemon how to spin up a new location - both the
    RepositoryLocationOrigin (which uniquely identifies the location) and any additional information
    that might be needed to create the user code deployment (for example, the docker image
    to use). The daemon then populates the entry with repository information or an error message after
    spinning the location up."""

    def __new__(
        cls,
        location_name,
        deployment_metadata,
        load_repositories_response=None,
        serialized_error_info=None,
        timestamp=None,
        data_updated_timestamp=None,
    ):
        return super(DagsterCloudWorkspaceEntry, cls).__new__(
            cls,
            check.str_param(location_name, "location_name"),
            check.opt_inst_param(
                deployment_metadata, "deployment_metadata", CodeDeploymentMetadata
            ),
            check.opt_inst_param(
                load_repositories_response, "load_repositories_response", LoadRepositoriesResponse
            ),
            check.opt_inst_param(
                serialized_error_info, "serialized_error_info", SerializableErrorInfo
            ),
            check.opt_float_param(
                timestamp,
                "timestamp",
            ),
            check.opt_float_param(
                data_updated_timestamp,
                "data_updated_timestamp",
            ),
        )

    def with_load_repositories_response(self, load_repositories_response):
        check.inst_param(
            load_repositories_response, "load_repositories_response", LoadRepositoriesResponse
        )
        return self._replace(
            load_repositories_response=load_repositories_response,
            serialized_error_info=None,
        )

    def with_serialized_error_info(self, serialized_error_info):
        check.inst_param(serialized_error_info, "serialized_error_info", SerializableErrorInfo)
        return self._replace(
            load_repositories_response=None,
            serialized_error_info=serialized_error_info,
        )

    def with_timestamp(self, timestamp):
        check.float_param(timestamp, "timestamp")
        return self._replace(
            timestamp=timestamp,
        )

    @property
    def outdated_data(self):
        return self.data_updated_timestamp and self.timestamp > self.data_updated_timestamp

    @property
    def status(self):
        outdated_data = self.timestamp and self.outdated_data
        if not outdated_data and self.load_repositories_response:
            return DagsterCloudWorkspaceEntryStatus.LOADED
        elif not outdated_data and self.serialized_error_info:
            return DagsterCloudWorkspaceEntryStatus.ERRORED
        else:
            return DagsterCloudWorkspaceEntryStatus.LOADING

    def get_display_metadata(self):
        metadata = {
            "python_file": self.deployment_metadata.python_file,
            "package_name": self.deployment_metadata.package_name,
            "image": self.deployment_metadata.image,
        }
        return {key: value for key, value in metadata.items() if value is not None}


@whitelist_for_serdes
class DagsterCloudApi(Enum):
    CHECK_FOR_WORKSPACE_UPDATES = "CHECK_FOR_WORKSPACE_UPDATES"
    LOAD_REPOSITORIES = "LOAD_REPOSITORIES"
    GET_EXTERNAL_EXECUTION_PLAN = "GET_EXTERNAL_EXECUTION_PLAN"
    GET_SUBSET_EXTERNAL_PIPELINE_RESULT = "GET_SUBSET_EXTERNAL_PIPELINE_RESULT"
    GET_EXTERNAL_PARTITION_CONFIG = "GET_EXTERNAL_PARTITION_CONFIG"
    GET_EXTERNAL_PARTITION_TAGS = "GET_EXTERNAL_PARTITION_TAGS"
    GET_EXTERNAL_PARTITION_NAMES = "GET_EXTERNAL_PARTITION_NAMES"
    GET_EXTERNAL_PARTITION_SET_EXECUTION_PARAM_DATA = (
        "GET_EXTERNAL_PARTITION_SET_EXECUTION_PARAM_DATA"
    )
    GET_EXTERNAL_SCHEDULE_EXECUTION_DATA = "GET_EXTERNAL_SCHEDULE_EXECUTION_DATA"
    GET_EXTERNAL_SENSOR_EXECUTION_DATA = "GET_EXTERNAL_SENSOR_EXECUTION_DATA"

    LAUNCH_RUN = "LAUNCH_RUN"
    LAUNCH_STEP = "LAUNCH_STEP"
    CHECK_STEP_HEALTH = "CHECK_STEP_HEALTH"
    TERMINATE_STEP = "TERMINATE_STEP"


@whitelist_for_serdes
class DagsterCloudApiSuccess(namedtuple("_DagsterCloudApiSuccess", "")):
    pass


@whitelist_for_serdes
class CheckForWorkspaceUpdatesSuccess(namedtuple("_CheckForWorkspaceUpdatesSuccess", "")):
    pass


@whitelist_for_serdes
class LoadRepositoriesArgs(namedtuple("_LoadRepositoryArgs", "location_origin")):
    def __new__(cls, location_origin):
        return super(cls, LoadRepositoriesArgs).__new__(
            cls,
            check.inst_param(location_origin, "location_origin", RepositoryLocationOrigin),
        )


@whitelist_for_serdes
class DagsterCloudRepositoryData(
    namedtuple("_DagsterCloudRepositoryData", "repo_name code_pointer external_repository_data")
):
    def __new__(cls, repo_name, code_pointer, external_repository_data):
        return super(cls, DagsterCloudRepositoryData).__new__(
            cls,
            check.str_param(repo_name, "repo_name"),
            check.inst_param(code_pointer, "code_pointer", CodePointer),
            check.inst_param(
                external_repository_data,
                "external_repository_data",
                ExternalRepositoryData,
            ),
        )


@whitelist_for_serdes
class LoadRepositoriesResponse(
    namedtuple("_LoadRepositoriesResponse", "repository_datas container_image executable_path")
):
    def __new__(cls, repository_datas, container_image, executable_path):
        return super(cls, LoadRepositoriesResponse).__new__(
            cls,
            check.list_param(
                repository_datas,
                "repository_datas",
                of_type=DagsterCloudRepositoryData,
            ),
            check.opt_str_param(container_image, "container_image"),
            check.opt_str_param(executable_path, "executable_path"),
        )


@whitelist_for_serdes
class DagsterCloudApiErrorResponse(namedtuple("DagsterCloudApiErrorResponse", "error_infos")):
    def __new__(cls, error_infos):
        return super(cls, DagsterCloudApiErrorResponse).__new__(
            cls,
            check.list_param(error_infos, "error_infos", of_type=SerializableErrorInfo),
        )


@whitelist_for_serdes
class LaunchRunArgs(namedtuple("_LaunchRunArgs", "pipeline_run")):
    def __new__(cls, pipeline_run):
        return super(cls, LaunchRunArgs).__new__(
            cls,
            check.inst_param(pipeline_run, "pipeline_run", PipelineRun),
        )


@whitelist_for_serdes
class StepHandlerArgs(namedtuple("_StepHandlerArgs", "persisted_step_handler_context")):
    def __new__(cls, persisted_step_handler_context: PersistedDagsterCloudStepHandlerContext):
        return super(cls, StepHandlerArgs).__new__(
            cls,
            check.inst_param(
                persisted_step_handler_context,
                "persisted_step_handler_context",
                PersistedDagsterCloudStepHandlerContext,
            ),
        )


@whitelist_for_serdes
class DagsterCloudApiRequest(
    namedtuple("_DagsterCloudApiRequest", "request_id request_api request_args expire_at")
):
    def __new__(
        cls,
        request_id,
        request_api,
        request_args,
        expire_at=None,
    ):
        return super(cls, DagsterCloudApiRequest).__new__(
            cls,
            check.str_param(request_id, "request_id"),
            check.inst_param(request_api, "request_api", DagsterCloudApi),
            request_args,
            check.opt_float_param(
                expire_at,
                "expire_at",
                default=(
                    pendulum.now("UTC") + timedelta(milliseconds=DEFAULT_EXPIRATION_MILLISECONDS)
                ).timestamp(),
            ),
        )

    @property
    def is_expired(self):
        return pendulum.now("UTC").timestamp() > self.expire_at



import json
from typing import Any, Callable, Dict, Iterable, List, Optional, Set, Tuple, Union

from dagster import check
from dagster.core.errors import DagsterRunAlreadyExists, DagsterSnapshotDoesNotExist
from dagster.core.events import DagsterEvent
from dagster.core.execution.backfill import BulkActionStatus, PartitionBackfill
from dagster.core.snap import (
    ExecutionPlanSnapshot,
    PipelineSnapshot,
    create_execution_plan_snapshot_id,
    create_pipeline_snapshot_id,
)
from dagster.core.storage.pipeline_run import PipelineRun, PipelineRunsFilter, RunRecord
from dagster.core.storage.runs.base import RunStorage
from dagster.daemon.types import DaemonHeartbeat
from dagster.serdes import (
    ConfigurableClass,
    ConfigurableClassData,
    deserialize_json_to_dagster_namedtuple,
    serialize_dagster_namedtuple,
)
from dagster.utils import utc_datetime_from_timestamp

from ..errors import GraphQLStorageError
from .queries import (
    ADD_BACKFILL_MUTATION,
    ADD_DAEMON_HEARTBEAT_MUTATION,
    ADD_EXECUTION_PLAN_SNAPSHOT_MUTATION,
    ADD_PIPELINE_SNAPSHOT_MUTATION,
    ADD_RUN_MUTATION,
    ADD_RUN_TAGS_MUTATION,
    GET_BACKFILLS_QUERY,
    GET_BACKFILL_QUERY,
    GET_DAEMON_HEARTBEATS_QUERY,
    GET_EXECUTION_PLAN_SNAPSHOT_QUERY,
    GET_PIPELINE_SNAPSHOT_QUERY,
    GET_RUNS_COUNT_QUERY,
    GET_RUNS_QUERY,
    GET_RUN_BY_ID_QUERY,
    GET_RUN_GROUPS_QUERY,
    GET_RUN_GROUP_QUERY,
    GET_RUN_RECORDS_QUERY,
    GET_RUN_TAGS_QUERY,
    HANDLE_RUN_EVENT_MUTATION,
    HAS_EXECUTION_PLAN_SNAPSHOT_QUERY,
    HAS_PIPELINE_SNAPSHOT_QUERY,
    HAS_RUN_QUERY,
    UPDATE_BACKFILL_MUTATION,
)


def _get_filters_input(filters) -> Optional[Dict[str, Any]]:
    filters = check.opt_inst_param(filters, "filters", PipelineRunsFilter)
    if filters is None:
        return None
    return {
        "runIds": filters.run_ids,
        "pipelineName": filters.pipeline_name,
        "statuses": [status.value for status in filters.statuses],
        "tags": [
            {"key": tag_key, "value": tag_value} for tag_key, tag_value in filters.tags.items()
        ]
        if filters.tags
        else [],
        "snapshotId": filters.snapshot_id,
        "updatedAfter": filters.updated_after.isoformat() if filters.updated_after else None,
    }


def _run_record_from_graphql(graphene_run_record: Dict) -> RunRecord:
    check.dict_param(graphene_run_record, "graphene_run_record")
    return RunRecord(
        storage_id=check.int_elem(graphene_run_record, "storageId"),
        pipeline_run=check.inst(
            deserialize_json_to_dagster_namedtuple(
                check.str_elem(graphene_run_record, "serializedPipelineRun")
            ),
            PipelineRun,
        ),
        create_timestamp=utc_datetime_from_timestamp(
            check.float_elem(graphene_run_record, "createTimestamp")
        ),
        update_timestamp=utc_datetime_from_timestamp(
            check.float_elem(graphene_run_record, "updateTimestamp")
        ),
    )


class GraphQLRunStorage(RunStorage, ConfigurableClass):
    def __init__(self, inst_data: ConfigurableClassData = None, override_graphql_client=None):
        """Initialize this class directly only for test (using `override_graphql_client`).
        Use the ConfigurableClass machinery to init from instance yaml."""
        self._inst_data = check.opt_inst_param(inst_data, "inst_data", ConfigurableClassData)
        self._override_graphql_client = override_graphql_client

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {}

    @staticmethod
    def from_config_value(inst_data: ConfigurableClassData, config_value):
        return GraphQLRunStorage(inst_data=inst_data)

    @property
    def _graphql_client(self):
        return (
            self._override_graphql_client
            if self._override_graphql_client
            else self._instance.graphql_client
        )

    def _execute_query(self, query, variables=None):
        res = self._graphql_client.execute(query, variable_values=variables)
        if "errors" in res:
            raise GraphQLStorageError(res)
        return res

    def add_run(self, pipeline_run: PipelineRun):
        check.inst_param(pipeline_run, "pipeline_run", PipelineRun)
        res = self._execute_query(
            ADD_RUN_MUTATION,
            variables={"serializedPipelineRun": serialize_dagster_namedtuple(pipeline_run)},
        )

        result = res["data"]["runs"]["addRun"]
        error = result.get("error")

        # Special-case some errors to match the RunStorage API
        if error:
            if error["className"] == "DagsterRunAlreadyExists":
                raise DagsterRunAlreadyExists(error["message"])
            if error["className"] == "DagsterSnapshotDoesNotExist":
                raise DagsterSnapshotDoesNotExist(error["message"])
            else:
                raise GraphQLStorageError(res)

        return pipeline_run

    def handle_run_event(self, run_id: str, event: DagsterEvent):
        check.inst_param(event, "event", DagsterEvent)
        self._execute_query(
            HANDLE_RUN_EVENT_MUTATION,
            variables={
                "runId": check.str_param(run_id, "run_id"),
                "serializedEvent": serialize_dagster_namedtuple(event),
            },
        )

    def get_runs(
        self, filters: PipelineRunsFilter = None, cursor: str = None, limit: int = None
    ) -> Iterable[PipelineRun]:
        res = self._execute_query(
            GET_RUNS_QUERY,
            variables={
                "filters": _get_filters_input(filters),
                "cursor": check.opt_str_param(cursor, "cursor"),
                "limit": check.opt_int_param(limit, "limit"),
            },
        )
        return [
            check.inst(deserialize_json_to_dagster_namedtuple(run), PipelineRun)
            for run in res["data"]["runs"]["getRuns"]
        ]

    def get_runs_count(self, filters: PipelineRunsFilter = None) -> int:
        res = self._execute_query(
            GET_RUNS_COUNT_QUERY,
            variables={
                "filters": _get_filters_input(filters),
            },
        )
        return res["data"]["runs"]["getRunsCount"]

    def get_run_group(self, run_id: str) -> Optional[Tuple[str, Iterable[PipelineRun]]]:
        res = self._execute_query(
            GET_RUN_GROUP_QUERY, variables={"runId": check.str_param(run_id, "run_id")}
        )
        run_group = res["data"]["runs"]["getRunGroup"]
        if run_group is None:
            return None

        return (
            run_group["rootRunId"],
            [
                check.inst(deserialize_json_to_dagster_namedtuple(run), PipelineRun)
                for run in run_group["serializedRuns"]
            ],
        )

    def get_run_groups(
        self, filters: PipelineRunsFilter = None, cursor: str = None, limit: int = None
    ) -> Dict[str, Dict[str, Union[Iterable[PipelineRun], int]]]:
        res = self._execute_query(
            GET_RUN_GROUPS_QUERY,
            variables={
                "filters": _get_filters_input(filters),
                "cursor": check.opt_str_param(cursor, "cursor"),
                "limit": check.opt_int_param(limit, "limit"),
            },
        )
        raw_run_groups = res["data"]["runs"]["getRunGroups"]

        run_groups = {}
        for run_group in raw_run_groups:
            run_groups[run_group["rootRunId"]] = {
                "runs": [
                    check.inst(deserialize_json_to_dagster_namedtuple(run), PipelineRun)
                    for run in run_group["serializedRuns"]
                ],
                "count": run_group["count"],
            }
        return run_groups

    def get_run_by_id(self, run_id: str) -> Optional[PipelineRun]:
        res = self._execute_query(
            GET_RUN_BY_ID_QUERY, variables={"runId": check.str_param(run_id, "run_id")}
        )
        run = res["data"]["runs"]["getRunById"]
        if run is None:
            return None

        return check.inst(deserialize_json_to_dagster_namedtuple(run), PipelineRun)

    def get_run_records(
        self,
        filters: PipelineRunsFilter = None,
        limit: int = None,
        order_by: str = None,
        ascending: bool = False,
    ) -> List[RunRecord]:
        res = self._execute_query(
            GET_RUN_RECORDS_QUERY,
            variables={
                "filters": _get_filters_input(filters),
                "limit": check.opt_int_param(limit, "limit"),
                "orderBy": check.opt_str_param(order_by, "order_by"),
                "ascending": check.opt_bool_param(ascending, "ascending"),
            },
        )
        return [_run_record_from_graphql(record) for record in res["data"]["runs"]["getRunRecords"]]

    def get_run_tags(self) -> List[Tuple[str, Set[str]]]:
        res = self._execute_query(GET_RUN_TAGS_QUERY)
        return [
            (run_tag["key"], set(run_tag["values"]))
            for run_tag in res["data"]["runs"]["getRunTags"]
        ]

    def add_run_tags(self, run_id: str, new_tags: Dict[str, str]):
        self._execute_query(
            ADD_RUN_TAGS_MUTATION,
            variables={
                "runId": check.str_param(run_id, "run_id"),
                "jsonNewTags": json.dumps(
                    check.dict_param(new_tags, "new_tags", key_type=str, value_type=str)
                ),
            },
        )

    def has_run(self, run_id: str) -> bool:
        res = self._execute_query(
            HAS_RUN_QUERY, variables={"runId": check.str_param(run_id, "run_id")}
        )
        return res["data"]["runs"]["hasRun"]

    def has_pipeline_snapshot(self, pipeline_snapshot_id: str) -> bool:
        res = self._execute_query(
            HAS_PIPELINE_SNAPSHOT_QUERY,
            variables={
                "pipelineSnapshotId": check.str_param(pipeline_snapshot_id, "pipeline_snapshot_id")
            },
        )
        return res["data"]["runs"]["hasPipelineSnapshot"]

    def add_pipeline_snapshot(self, pipeline_snapshot: PipelineSnapshot) -> str:
        self._execute_query(
            ADD_PIPELINE_SNAPSHOT_MUTATION,
            variables={
                "serializedPipelineSnapshot": serialize_dagster_namedtuple(
                    check.inst_param(pipeline_snapshot, "pipeline_snapshot", PipelineSnapshot)
                )
            },
        )
        return create_pipeline_snapshot_id(pipeline_snapshot)

    def get_pipeline_snapshot(self, pipeline_snapshot_id: str) -> PipelineSnapshot:
        res = self._execute_query(
            GET_PIPELINE_SNAPSHOT_QUERY,
            variables={
                "pipelineSnapshotId": check.str_param(pipeline_snapshot_id, "pipeline_snapshot_id")
            },
        )
        return check.inst(
            deserialize_json_to_dagster_namedtuple(res["data"]["runs"]["getPipelineSnapshot"]),
            PipelineSnapshot,
        )

    def has_execution_plan_snapshot(self, execution_plan_snapshot_id: str) -> bool:
        res = self._execute_query(
            HAS_EXECUTION_PLAN_SNAPSHOT_QUERY,
            variables={
                "executionPlanSnapshotId": check.str_param(
                    execution_plan_snapshot_id, "execution_plan_snapshot_id"
                )
            },
        )
        return res["data"]["runs"]["hasExecutionPlanSnapshot"]

    def add_execution_plan_snapshot(self, execution_plan_snapshot: ExecutionPlanSnapshot) -> str:
        self._execute_query(
            ADD_EXECUTION_PLAN_SNAPSHOT_MUTATION,
            variables={
                "serializedExecutionPlanSnapshot": serialize_dagster_namedtuple(
                    check.inst_param(
                        execution_plan_snapshot, "execution_plan_snapshot", ExecutionPlanSnapshot
                    )
                )
            },
        )
        return create_execution_plan_snapshot_id(execution_plan_snapshot)

    def get_execution_plan_snapshot(self, execution_plan_snapshot_id: str) -> ExecutionPlanSnapshot:
        res = self._execute_query(
            GET_EXECUTION_PLAN_SNAPSHOT_QUERY,
            variables={
                "executionPlanSnapshotId": check.str_param(
                    execution_plan_snapshot_id, "execution_plan_snapshot_id"
                )
            },
        )
        return check.inst(
            deserialize_json_to_dagster_namedtuple(res["data"]["runs"]["getExecutionPlanSnapshot"]),
            ExecutionPlanSnapshot,
        )

    def add_daemon_heartbeat(self, daemon_heartbeat: DaemonHeartbeat):
        self._execute_query(
            ADD_DAEMON_HEARTBEAT_MUTATION,
            variables={
                "serializedDaemonHeartbeat": serialize_dagster_namedtuple(
                    check.inst_param(daemon_heartbeat, "daemon_heartbeat", DaemonHeartbeat)
                )
            },
        )

    def build_missing_indexes(
        self, print_fn: Callable = lambda _: None, force_rebuild_all: bool = False
    ):
        raise Exception("Not allowed to build indexes from user cloud")

    def get_daemon_heartbeats(self) -> Dict[str, DaemonHeartbeat]:
        res = self._execute_query(GET_DAEMON_HEARTBEATS_QUERY)
        return {
            key: check.inst(deserialize_json_to_dagster_namedtuple(heartbeat), DaemonHeartbeat)
            for key, heartbeat in json.loads(res["data"]["runs"]["getDaemonHeartbeats"]).items()
        }

    def delete_run(self, run_id: str):
        raise Exception("Not allowed to delete runs from user cloud")

    def wipe_daemon_heartbeats(self):
        raise Exception("Not allowed to wipe heartbeats from user cloud")

    def wipe(self):
        raise Exception("Not allowed to wipe runs from user cloud")

    def has_bulk_actions_table(self) -> bool:
        return False

    def get_backfills(self, status: BulkActionStatus = None, cursor: str = None, limit: int = None):
        """Get a list of partition backfills"""
        res = self._execute_query(
            GET_BACKFILLS_QUERY,
            variables={
                "status": status.value if status else None,
                "cursor": check.opt_str_param(cursor, "cursor"),
                "limit": check.opt_int_param(limit, "limit"),
            },
        )
        return [
            check.inst(deserialize_json_to_dagster_namedtuple(backfill), PartitionBackfill)
            for backfill in res["data"]["runs"]["getBackfills"]
        ]

    def get_backfill(self, backfill_id: str) -> PartitionBackfill:
        """Get a single partition backfill"""
        res = self._execute_query(GET_BACKFILL_QUERY, variables={"backfillId": backfill_id})
        backfill = res["data"]["runs"]["getBackfill"]
        return check.inst(deserialize_json_to_dagster_namedtuple(backfill), PartitionBackfill)

    def add_backfill(self, partition_backfill: PartitionBackfill):
        """Add partition backfill to run storage"""
        self._execute_query(
            ADD_BACKFILL_MUTATION,
            variables={
                "serializedPartitionBackfill": serialize_dagster_namedtuple(partition_backfill)
            },
        )

    def update_backfill(self, partition_backfill: PartitionBackfill):
        """Update a partition backfill in run storage"""
        self._execute_query(
            UPDATE_BACKFILL_MUTATION,
            variables={
                "serializedPartitionBackfill": serialize_dagster_namedtuple(partition_backfill)
            },
        )



API_TOKEN_HEADER = "Dagster-Cloud-Api-Token"


def get_hardcoded_test_agent_token(organization_name):
    return f"agent-test_agent_token_{organization_name}"


def get_hardcoded_test_user_token(organization_name, user_name):
    return f"user-test_agent_token_{organization_name}_{user_name}"



import subprocess
import sys
import threading
from collections import namedtuple
from typing import List

from dagster import check
from dagster.core.events import DagsterEvent
from dagster.core.executor.step_delegating import StepHandler
from dagster.core.executor.step_delegating.step_handler.base import StepHandlerContext
from dagster.core.host_representation.grpc_server_registry import GrpcServerEndpoint
from dagster.core.launcher import DefaultRunLauncher
from dagster.core.types.loadable_target_origin import LoadableTargetOrigin
from dagster.daemon.daemon import get_default_daemon_logger
from dagster.grpc.client import DagsterGrpcClient, client_heartbeat_thread
from dagster.grpc.server import GrpcServerProcess
from dagster.serdes import ConfigurableClass, create_snapshot_id, serialize_dagster_namedtuple

from .user_code_launcher import ReconcileUserCodeLauncher


class ProcessUserCodeEntry(
    namedtuple(
        "_ProcessUserCodeEntry",
        "grpc_server_process grpc_client heartbeat_shutdown_event heartbeat_thread",
    )
):
    def __new__(cls, grpc_server_process, grpc_client, heartbeat_shutdown_event, heartbeat_thread):
        return super(ProcessUserCodeEntry, cls).__new__(
            cls,
            check.inst_param(grpc_server_process, "grpc_server_process", GrpcServerProcess),
            check.inst_param(grpc_client, "grpc_client", DagsterGrpcClient),
            check.inst_param(heartbeat_shutdown_event, "heartbeat_shutdown_event", threading.Event),
            check.inst_param(heartbeat_thread, "heartbeat_thread", threading.Thread),
        )


class ProcessUserCodeLauncher(ReconcileUserCodeLauncher, ConfigurableClass):
    def __init__(self, inst_data=None, graphql_client=None, wait_for_processes=False):
        self._inst_data = inst_data
        self._logger = get_default_daemon_logger("ProcessDagsterUserCodeLauncher")

        # Dict[int, ProcessUserCodeEntry], map from pid to all servers being spun up (including
        # old servers in the process of being shut down)
        self._process_entries = {}

        # Dict[str, int], map from hash of location_name + metadata (see _get_process_key)
        # to the pid for that location-metadata combination
        self._active_pids = {}

        self._heartbeat_ttl = 60
        self._wait_for_processes = wait_for_processes

        super(ProcessUserCodeLauncher, self).__init__(graphql_client=graphql_client)

        self._launcher = DefaultRunLauncher()

    def register_instance(self, instance):
        super().register_instance(instance)
        self._launcher.register_instance(instance)

    @property
    def inst_data(self):
        return self._inst_data

    @classmethod
    def config_type(cls):
        return {}

    @staticmethod
    def from_config_value(inst_data, config_value):
        return ProcessUserCodeLauncher(inst_data=inst_data)

    def _add_server(self, location_name, metadata):
        process_key = self._get_process_key(location_name, metadata)
        existing_pid = self._get_existing_pid(location_name, metadata)
        if existing_pid:
            self._remove_pid(existing_pid)
            del self._active_pids[process_key]

        return self._start_new_server(location_name, metadata)

    def _start_new_server(self, location_name, metadata):
        process_key = self._get_process_key(location_name, metadata)
        loadable_target_origin = self._get_loadable_target_origin(metadata)
        server_process = GrpcServerProcess(
            loadable_target_origin=loadable_target_origin,
            heartbeat=True,
            heartbeat_timeout=self._heartbeat_ttl,
        )
        server_id = self._wait_for_server(
            host="localhost", port=server_process.port, socket=server_process.socket
        )

        client = DagsterGrpcClient(
            port=server_process.port,
            socket=server_process.socket,
            host="localhost",
            use_ssl=False,
        )

        heartbeat_shutdown_event = threading.Event()
        heartbeat_thread = threading.Thread(
            target=client_heartbeat_thread,
            args=(client, heartbeat_shutdown_event),
        )
        heartbeat_thread.daemon = True
        heartbeat_thread.start()

        pid = server_process.pid

        self._process_entries[server_process.pid] = ProcessUserCodeEntry(
            server_process,
            client,
            heartbeat_shutdown_event,
            heartbeat_thread,
        )
        self._active_pids[process_key] = pid

        endpoint = GrpcServerEndpoint(
            server_id=server_id,
            host="localhost",
            port=server_process.port,
            socket=server_process.socket,
        )

        return endpoint

    def _get_process_key(self, location_name, metadata):
        return f"{location_name}_{create_snapshot_id(metadata)[0:6]}"

    def _get_loadable_target_origin(self, metadata):
        return LoadableTargetOrigin(
            executable_path=sys.executable,
            python_file=metadata.python_file,
            package_name=metadata.package_name,
        )

    def _get_existing_pid(self, location_name, metadata):
        process_key = self._get_process_key(location_name, metadata)
        return self._active_pids.get(process_key)

    def _gen_update_server(self, location_name, old_metadata, new_metadata):
        old_process_key = self._get_process_key(location_name, old_metadata)
        new_process_key = self._get_process_key(location_name, new_metadata)
        existing_pid = self._get_existing_pid(location_name, old_metadata)

        updated_server = self._start_new_server(location_name, new_metadata)
        yield updated_server

        if existing_pid:
            self._logger.info(
                "Stopping old process for location {location_name}".format(
                    location_name=location_name,
                )
            )
            self._remove_pid(existing_pid)
            if old_process_key != new_process_key:
                del self._active_pids[old_process_key]

    def _remove_server(self, location_name, metadata):
        process_key = self._get_process_key(location_name, metadata)
        existing_pid = self._get_existing_pid(location_name, metadata)
        if existing_pid:
            self._logger.info(
                "Stopping process for location {location_name}".format(
                    location_name=location_name,
                )
            )
            self._remove_pid(existing_pid)
            del self._active_pids[process_key]

    def _remove_pid(self, pid):
        process_entry = self._process_entries[pid]
        process_entry.heartbeat_shutdown_event.set()
        process_entry.heartbeat_thread.join()
        # Rely on heartbeat failure to eventually kill the process
        del self._process_entries[pid]

    def step_handler(self):
        return ProcessStepHandler()

    def run_launcher(self):
        return self._launcher

    def __exit__(self, exception_type, exception_value, traceback):
        self._launcher.dispose()
        while len(self._process_entries):
            pid = next(iter(self._process_entries))
            process_entry = self._process_entries[pid]

            self._remove_pid(pid)
            if self._wait_for_processes:
                process_entry.grpc_client.shutdown_server()
                process_entry.grpc_server_process.wait()


class ProcessStepHandler(StepHandler):
    @property
    def name(self) -> str:
        return "ProcessStepHandler"

    def launch_step(self, step_handler_context: StepHandlerContext) -> List[DagsterEvent]:
        input_json = serialize_dagster_namedtuple(step_handler_context.execute_step_args)
        args = ["dagster", "api", "execute_step", input_json]
        subprocess.Popen(args)
        return []

    def check_step_health(self, step_handler_context: StepHandlerContext) -> List[DagsterEvent]:
        # TODO not implemented
        return []

    def terminate_step(self, step_handler_context: StepHandlerContext) -> List[DagsterEvent]:
        # TODO not implemented
        return []
